import argparse
import sys
import yaml

from . import builtins
from . import context
from . import util
from . import registry

parser = argparse.ArgumentParser(prog="y4")
parser.add_argument("input")
parser.add_argument(
    "-p",
    "--path",
    action="append",
    default=[],
    help="add PATH to the module search path",
)
parser.add_argument(
    "--opt",
    nargs=2,
    action="append",
    default=[],
    metavar=("KEY", "VALUE"),
    help="pass the (KEY, VALUE) pair as an option",
)
parser.add_argument(
    "--trace",
    required=False,
    action="store_true",
    default=False,
    help="enable tracing to stderr",
)


def main():
    args = parser.parse_args()

    env = context.Environment()
    for k, v in args.opt:
        env.set_option(k, v)
    for path in args.path:
        env.add_search_path(path)

    ctx = context.Context(env=env)
    if args.trace:
        ctx.enable_tracing = True

    with open(args.input) as f:
        roots = yaml.compose_all(f, Loader=util.YamlLoader)
        for root in roots:
            if root.tag == "tag:y4.managarm.org:preamble":
                d = ctx.assemble_dict_keys(root)
                for tag, binding in context.process_bindings(ctx, d):
                    ctx.bind(tag, binding)
            else:
                out = ctx.evaluate(root)
                sys.stdout.write(
                    yaml.dump(
                        out,
                        sort_keys=False,
                        explicit_start=True,
                        Dumper=util.YamlDumper,
                    )
                )


if __name__ == "__main__":
    main()



import os
import pytest
import subprocess
import sys
import yaml

cases = [
    "map.y4.yml",
    "reduce.y4.yml",
]


@pytest.mark.parametrize("path", cases)
def test_y4(path):
    args = [sys.executable, "-m", "y4", os.path.join("tests", path)]
    result = subprocess.run(args, stdout=subprocess.PIPE)
    assert result.returncode == 0

    yml = yaml.load(result.stdout, Loader=yaml.SafeLoader)
    assert yml["result"] == yml["expected"]



import yaml

YAML_SEQ_TAG = "tag:yaml.org,2002:seq"
YAML_MAP_TAG = "tag:yaml.org,2002:map"
YAML_STR_TAG = "tag:yaml.org,2002:str"

try:
    YamlLoader = yaml.CSafeLoader
except AttributeError:
    YamlLoader = yaml.SafeLoader

try:
    YamlDumper = yaml.CSafeDumper
except AttributeError:
    YamlDumper = yaml.SafeDumper


class InternalNode(yaml.Node):
    id = "internal"

    def __init__(self, tag, value):
        self.tag = tag
        self.value = value


class Y4Error(Exception):
    pass


def validate_node(node, desc, *, kind=None, tag=None):
    if kind is not None and not isinstance(node, kind):
        raise Y4Error(
            f"Invalid node kind {type(node).id} for {desc}, expected {kind.id}"
        )
    if tag is not None and node.tag != tag:
        raise Y4Error(f"Invalid node tag {node.tag} for {desc}, expected {tag}")


def get_local(tag):
    if not tag.startswith("!"):
        raise Y4Error(f"unexpected tag {tag}")
    return tag[1:]


def get_marker_tag(node):
    if not isinstance(node, yaml.ScalarNode):
        raise Y4Error("Marker nodes must be scalar")
    if node.value:
        raise Y4Error("Marker nodes must be empty")
    return node.tag


def copy_node(node, *, tag=None):
    if tag is None:
        tag = node.tag

    if isinstance(node, yaml.SequenceNode):
        return yaml.SequenceNode(tag, node.value)
    elif isinstance(node, yaml.MappingNode):
        return yaml.MappingNode(tag, node.value)
    elif isinstance(node, yaml.ScalarNode):
        return yaml.ScalarNode(tag, node.value)
    else:
        raise Y4Error(f"Unknown kind when transmuting tag")


def construct(node, *, tag=None):
    if tag is None:
        tag = node.tag

    if tag == "tag:yaml.org,2002:seq":
        if not isinstance(node, yaml.SequenceNode):
            raise Y4Error(f"Expected sequence node for {tag}")
        return [construct(item) for item in node.value]
    elif tag == "tag:yaml.org,2002:map":
        if not isinstance(node, yaml.MappingNode):
            raise Y4Error(f"Expected mapping node for {tag}")
        return {construct(k): construct(v) for k, v in node.value}
    elif tag == "tag:yaml.org,2002:str":
        if not isinstance(node, yaml.ScalarNode):
            raise Y4Error(f"Expected scalar node for {tag}")
        return node.value
    elif tag == "tag:yaml.org,2002:bool":
        if not isinstance(node, yaml.ScalarNode):
            raise Y4Error(f"Expected scalar node for {tag}")
        if node.value == "true":
            return True
        elif node.value == "false":
            return False
        else:
            raise Y4Error(f"Unexpected value {node.value} for boolean")
    elif tag == "tag:yaml.org,2002:int":
        if not isinstance(node, yaml.ScalarNode):
            raise Y4Error(f"Expected scalar node for {tag}")
        return int(node.value)
    else:
        raise Y4Error(f"Cannot construct value from tag {tag}")


def represent(data):
    if isinstance(data, dict):
        value = []
        for k, v in data.items():
            rep = (represent(k), represent(v))
            value.append(rep)
        return yaml.MappingNode("tag:yaml.org,2002:map", value)
    elif isinstance(data, list):
        value = []
        for item in data:
            rep = represent(item)
            value.append(rep)
        return yaml.SequenceNode("tag:yaml.org,2002:seq", value)
    elif isinstance(data, str):
        return yaml.ScalarNode("tag:yaml.org,2002:str", data)
    elif isinstance(data, bool):
        if data is True:
            value = "true"
        elif data is False:
            value = "false"
        else:
            raise Y4Error(f"Neither true or false: {data}")
        return yaml.ScalarNode("tag:yaml.org,2002:bool", value)
    elif isinstance(data, int):
        return yaml.ScalarNode("tag:yaml.org,2002:int", str(data))
    else:
        raise Y4Error(f"Cannot represent value of type {type(data)}")



from . import context

_builtin_registry = {}


def get_builtin_modules():
    return _builtin_registry


class BuiltinRule(context.Rule):
    def __init__(self, func):
        self.normalize = func


# Function decorator that registers a builtin.
def builtin(*, tag):
    def decorate(func):
        prefix, suffix = tag.split("::")

        module = _builtin_registry.get(prefix)
        if module is None:
            module = context.Namespace()
            _builtin_registry[prefix] = module

        rule = BuiltinRule(func)
        module._bindings[suffix] = rule

        return func

    return decorate



import functools
import operator
import yaml

from .. import context
from .. import util
from ..registry import builtin


class Y4Function:
    def __init__(self, inner_ctx, args, expr):
        self._inner_ctx = inner_ctx
        self._args = args
        self._expr = expr

    def apply(self, *nodes):
        # Note that the arguments are not normalized before application.
        # They will usually be normalized by the caller.

        # We compute the function using the inner context with !arg introduced.

        apply_ctx = context.Context(self._inner_ctx)

        if len(self._args) != len(nodes):
            raise util.Y4Error(
                f"y4 function that takes {len(self._args)} argument(s)"
                f" was called with {len(nodes)} argument(s)"
            )
        for tag, node in zip(self._args, nodes):
            apply_ctx.bind(tag, context.ConstRule(node))

        return apply_ctx.normalize(self._expr)


@builtin(tag="std::let")
def let(ctx, node):
    d = ctx.assemble_dict_keys(node)
    # TODO: We should not assume that d["const"] is a MappingNode.
    nested_ctx = context.Context(ctx)
    for tag, rule in context.process_bindings(ctx, d):
        nested_ctx.bind(tag, rule)
    res = nested_ctx.normalize(d["in"])
    return res


@builtin(tag="std::fn")
def fn(ctx, node):
    # Note that the inner context of the function is a snapshot of the current context.
    # It is not equal to the current context, i.e., subsequently introduced rules
    # will not affect the inner context.
    inner_ctx = context.Context(ctx)

    d = ctx.assemble_dict_keys(node)

    if "arg" in d and "args" in d:
        raise util.Y4Error("std::fn cannot specify both arg: and args:")
    elif "arg" in d:
        args = [util.get_local(util.get_marker_tag(d["arg"]))]
    elif "args" in d:
        util.validate_node(
            d["args"], "std::fn", kind=yaml.SequenceNode, tag=util.YAML_SEQ_TAG
        )
        args = [util.get_local(util.get_marker_tag(arg)) for arg in d["args"].value]
    else:
        raise util.Y4Error("std::fn has neither arg: nor args:")

    expr = d["return"]
    return util.InternalNode(
        "tag:y4.managarm.org:function", Y4Function(inner_ctx, args, expr)
    )


@builtin(tag="std::normalize")
def normalize(ctx, node):
    tf = ctx.normalize(node, tag="tag:yaml.org,2002:map")
    d = ctx.assemble_dict_keys(tf)
    util.validate_node(
        d["ctx"],
        "ctx: parameter of std::normalize",
        kind=util.InternalNode,
        tag="tag:y4.managarm.org:context",
    )
    inner_ctx = d["ctx"].value
    return inner_ctx.normalize(d["node"])


@builtin(tag="std::opt")
def opt(ctx, node):
    k = ctx.evaluate(node, tag="tag:yaml.org,2002:str")
    return util.represent(ctx.env.get_option(k))


@builtin(tag="std::ite")
def ite(ctx, node):
    d = ctx.assemble_dict_keys(node)
    cond = ctx.evaluate(d["if"])
    if cond:
        return ctx.normalize(d["then"])
    else:
        return ctx.normalize(d["else"])


@builtin(tag="std::get")
def get(ctx, node):
    if isinstance(node, yaml.SequenceNode):
        raw_obj = node.value[0]
        raw_key = node.value[1]
    elif isinstance(node, yaml.MappingNode):
        d = ctx.assemble_dict_keys(node)
        raw_obj = d["object"]
        raw_key = d["key"]
    else:
        raise util.Y4Error("Invalid node kind for !std::get")
    obj = ctx.normalize(raw_obj)
    key = ctx.evaluate(raw_key)
    obj_dict = ctx.assemble_dict_keys(obj)
    return obj_dict[key]


@builtin(tag="std::contains")
def contains(ctx, node):
    obj = ctx.evaluate(node, tag="tag:yaml.org,2002:map")
    return util.represent(obj["item"] in obj["list"])


@builtin(tag="std::apply")
def apply(ctx, node):
    tf = ctx.normalize(node, tag="tag:yaml.org,2002:map")
    d = ctx.assemble_dict_keys(tf)

    # Extract fn:
    util.validate_node(
        d["fn"],
        "fn: parameter for std::apply",
        kind=util.InternalNode,
        tag="tag:y4.managarm.org:function",
    )
    func = d["fn"].value

    # Extract either arg: or args:
    if "arg" in d and "args" in d:
        raise util.Y4Error("std::apply cannot specify both arg: and args:")
    elif "arg" in d:
        args = [d["arg"]]
    elif "args" in d:
        util.validate_node(
            d["args"], "std::apply", kind=yaml.SequenceNode, tag=util.YAML_SEQ_TAG
        )
        args = d["args"].value
    else:
        raise util.Y4Error("std::apply has neither arg: nor args:")

    return func.apply(*args)


@builtin(tag="std::splice_if")
def splice_if(ctx, node):
    value = []
    for raw_item in node.value:
        item = ctx.normalize(raw_item)
        d = ctx.assemble_dict_keys(item)
        if ctx.evaluate(d["if"]):
            value.append(d["item"])

    return yaml.SequenceNode("tag:y4.managarm.org:splice", value)


# List manipulation.


@builtin(tag="std::map")
def map(ctx, node):
    tf = ctx.normalize(node, tag=util.YAML_MAP_TAG)
    d = ctx.assemble_dict_keys(tf)
    util.validate_node(
        d["list"],
        "list: parameter of std::map",
        kind=yaml.SequenceNode,
        tag=util.YAML_SEQ_TAG,
    )
    util.validate_node(
        d["fn"],
        "fn: parameter of std::map",
        kind=util.InternalNode,
        tag="tag:y4.managarm.org:function",
    )
    fn = d["fn"].value
    return yaml.SequenceNode(
        util.YAML_SEQ_TAG, [fn.apply(item) for item in d["list"].value]
    )


@builtin(tag="std::reduce")
def reduce(ctx, node):
    tf = ctx.normalize(node, tag=util.YAML_MAP_TAG)
    d = ctx.assemble_dict_keys(tf)
    util.validate_node(
        d["list"],
        "list: parameter of std::reduce",
        kind=yaml.SequenceNode,
        tag=util.YAML_SEQ_TAG,
    )
    util.validate_node(
        d["fn"],
        "fn: parameter of std::reduce",
        kind=util.InternalNode,
        tag="tag:y4.managarm.org:function",
    )
    fn = d["fn"].value
    return functools.reduce(fn.apply, d["list"].value, d["init"])


# String manipulation.


@builtin(tag="std::join")
def join(ctx, node):
    parts = [ctx.evaluate(item) for item in node.value]

    return util.represent("".join(parts))


# Arithmetic.


def extract_operands(ctx, node, desc):
    if isinstance(node, yaml.SequenceNode):
        tf = ctx.normalize(node, tag=util.YAML_SEQ_TAG)
        return tf.value
    elif isinstance(node, yaml.MappingNode):
        tf = ctx.normalize(node, tag=util.YAML_MAP_TAG)
        d = ctx.assemble_dict_keys(tf)
        util.validate_node(
            d["ops"], desc, kind=yaml.SequenceNode, tag=util.YAML_SEQ_TAG
        )
        return d["ops"].value
    else:
        raise util.Y4Error(f"Invalid node kind for {desc}")


@builtin(tag="std::add")
def add(ctx, node):
    ops = extract_operands(ctx, node, "std::add")
    return util.represent(sum(util.construct(op) for op in ops))


@builtin(tag="std::mul")
def mul(ctx, node):
    ops = extract_operands(ctx, node, "std::mul")
    it = (util.construct(op) for op in ops)
    return util.represent(functools.reduce(operator.mul, it, 1))


@builtin(tag="std::sub")
def sub(ctx, node):
    ops = extract_operands(ctx, node, "std::sub")
    if len(ops) != 2:
        raise util.Y4Error("Expected exactly two operands for std::sub")
    lhs = util.construct(ops[0])
    rhs = util.construct(ops[1])
    return util.represent(lhs - rhs)


@builtin(tag="std::div")
def div(ctx, node):
    ops = extract_operands(ctx, node, "std::div")
    if len(ops) != 2:
        raise util.Y4Error("Expected exactly two operands for std::div")
    lhs = util.construct(ops[0])
    rhs = util.construct(ops[1])
    return util.represent(lhs // rhs)


@builtin(tag="std::mod")
def mod(ctx, node):
    ops = extract_operands(ctx, node, "std::mod")
    if len(ops) != 2:
        raise util.Y4Error("Expected exactly two operands for std::mod")
    lhs = util.construct(ops[0])
    rhs = util.construct(ops[1])
    return util.represent(lhs % rhs)



import jq

from .. import util
from ..registry import builtin


class JqFunction:
    def __init__(self, script):
        self.program = jq.compile(script)

    def apply(self, ctx, node):
        obj = ctx.evaluate(node, tag="tag:yaml.org,2002:map")
        results = self.program.input_value(obj).all()
        if not results:
            raise util.Y4Error("jq script returned no results")
        if len(results) > 1:
            raise util.Y4Error("jq script returned more than one result")
        return util.represent(results[0])


@builtin(tag="jq::fn")
def fn(ctx, node):
    script = node.value
    return util.InternalNode("tag:y4.managarm.org:function", JqFunction(script))



from copy import copy
import enum
import functools
import os
import sys
import yaml

from . import util
from . import registry


class Binding:
    pass


class Namespace(Binding):
    def __init__(self):
        self._bindings = {}

    def bind(self, tag, binding):
        assert tag not in self._bindings.keys()
        self._bindings[tag] = binding

    def lookup(self, tag):
        return self._bindings.get(tag)


class Rule(Binding):
    pass


class Environment:
    def __init__(self):
        self._mods = {}
        self._search_paths = []
        self._opts = {}

    def set_option(self, k, v):
        self._opts[k] = v

    def get_option(self, k):
        return self._opts.get(k)

    def add_search_path(self, path):
        self._search_paths.append(path)

    def locate_module_loaders(self, name):
        module = registry.get_builtin_modules().get(name)
        if module:
            yield functools.partial(self._load_builtin, module)

        for search_path in self._search_paths:
            path = os.path.join(search_path, f"{name}.yaml")
            if os.path.exists(path):
                yield functools.partial(self._load_from_path, path)

    def get_or_load_module(self, name):
        loaded = self._mods.get(name)
        if loaded:
            return loaded

        loaders = list(self.locate_module_loaders(name))
        if not loaders:
            raise util.Y4Error(f"Module {name} does not exist")
        if len(loaders) > 1:
            raise util.Y4Error(f"Module {name} is ambiguous")

        module = loaders[0]()
        self._mods[name] = module
        return module

    def _load_builtin(self, module):
        return module

    def _load_from_path(self, path):
        with open(path) as f:
            root = yaml.compose(f, Loader=util.YamlLoader)

        ctx = Context(env=self)
        module = Namespace()

        d = ctx.assemble_dict_keys(root)
        for tag, binding in process_bindings(ctx, d):
            module.bind(tag, binding)

        return module


class Context:
    def __init__(self, parent=None, *, env=None):
        if parent is None:
            assert env is not None

            self.env = env
            self.enable_tracing = False
            self._bindings = {}
        else:
            assert env is None

            self.env = parent.env
            self.enable_tracing = parent.enable_tracing
            self._bindings = copy(parent._bindings)

    # Binds a tag to a Rule or Namespace.
    def bind(self, tag, binding):
        if not isinstance(binding, Binding):
            raise RuntimeError("Bindings need to inherit from Binding class")
        if tag in self._bindings:
            raise util.Y4Error(f"Duplicate tag definition: {tag} is already defined")
        self._bindings[tag] = binding

    # Resolves a qualified tag or returns None on error.
    def resolve(self, tag):
        components = tag.split("::")
        if not components:
            return None
        first, *tail = components

        # Look up the first component.
        current = self._bindings.get(first)
        if current is None:
            return None

        # Look up nested components.
        for follow in tail:
            current = current.lookup(follow)
            if current is None:
                return None

        assert current is not None
        return current

    def normalize(self, node, *, tag=None):
        self.trace("normalize", node.tag, node)

        if tag is None:
            tag = node.tag

        if tag.startswith("!"):
            suffix = tag[1:]

            tf = self.resolve(suffix)
            if tf is not None:
                return tf.normalize(self, node)
            else:
                raise util.Y4Error(f"Undefined binding !{suffix}")
        elif tag == "tag:yaml.org,2002:seq":
            if not isinstance(node, yaml.SequenceNode):
                raise util.Y4Error(f"Invalid node kind for {tag}")
            value = []
            for item in node.value:
                tf = self.normalize(item)
                if tf.tag == "tag:y4.managarm.org:splice":
                    value.extend(tf.value)
                else:
                    value.append(tf)
            return yaml.SequenceNode(tag, value)
        elif tag == "tag:yaml.org,2002:map":
            if not isinstance(node, yaml.MappingNode):
                raise util.Y4Error(f"Invalid node kind for {tag}")
            value = []
            for k, v in node.value:
                tf = (self.normalize(k), self.normalize(v))
                value.append(tf)
            return yaml.MappingNode(tag, value)
        elif tag in [
            "tag:yaml.org,2002:str",
            "tag:yaml.org,2002:int",
            "tag:yaml.org,2002:bool",
        ]:
            if not isinstance(node, yaml.ScalarNode):
                raise util.Y4Error(f"Invalid node kind for {tag}")
            return yaml.ScalarNode(tag, node.value)
        elif tag in [
            "tag:y4.managarm.org:function",
            "tag:y4.managarm.org:context",
        ]:
            if not isinstance(node, util.InternalNode):
                raise util.Y4Error(f"Invalid node kind for {tag}")
            return util.InternalNode(tag, node.value)
        else:
            raise util.Y4Error(f"Unexpected YAML tag {node.tag}")

    def evaluate(self, node, *, tag=None):
        self.trace("evaluate", node)
        tf = self.normalize(node, tag=tag)
        return util.construct(tf)

    def assemble_dict_keys(self, node):
        d = {}
        if isinstance(node, yaml.MappingNode):
            for k, v in node.value:
                d[self.evaluate(k)] = v
        else:
            raise util.Y4Error(f"Cannot assemble dict from {type(node)}")
        return d

    def trace(self, event, *args):
        if not self.enable_tracing:
            return
        print(event, *args, file=sys.stderr)


class ConstRule(Rule):
    def __init__(self, const):
        self.const = const

    def normalize(self, ctx, node):
        return self.const


class CustomRule(Rule):
    def __init__(self, fn, *, normalize):
        self._fn = fn
        self._normalize = normalize

    def normalize(self, ctx, node):
        # Custom rules normalize the node as if it had the default tag,
        # i.e., the tag that the non-specific tag ! would resolve to.
        if isinstance(node, yaml.SequenceNode):
            base_tag = "tag:yaml.org,2002:seq"
        elif isinstance(node, yaml.MappingNode):
            base_tag = "tag:yaml.org,2002:map"
        elif isinstance(node, yaml.ScalarNode):
            # Scalars are normalized as strings.
            base_tag = "tag:yaml.org,2002:str"
        else:
            raise util.Y4Error(f"Cannot apply custom rule to node kind {type(node)}")

        if self._normalize:
            tf = ctx.normalize(node, tag=base_tag)
            return self._fn.apply(tf)
        else:
            ctx_node = util.InternalNode("tag:y4.managarm.org:context", ctx)
            tf = util.copy_node(node, tag=base_tag)
            return self._fn.apply(ctx_node, tf)


def process_bindings(ctx, d):
    if "import" in d:
        # TODO: Validate that d["import"] has the right (kind, tag).
        for item in d["import"].value:
            tag = util.get_local(item.tag)
            module = ctx.env.get_or_load_module(tag)
            ctx.bind(tag, module)
    if "const" in d:
        # TODO: Validate that d["const"] has the right (kind, tag).
        for k, v in d["const"].value:
            tag = util.get_local(k.tag)
            tf = ctx.normalize(v)
            yield (tag, ConstRule(tf))
    if "rule" in d:
        # TODO: Validate that d["rule"] has the right (kind, tag).
        for k, v in d["rule"].value:
            tag = util.get_local(k.tag)
            tf = ctx.normalize(v)
            if tf.tag == "tag:y4.managarm.org:function":
                util.validate_node(tf, "custom rule", kind=util.InternalNode)
                fn = tf.value
                normalize = True
            elif tf.tag == util.YAML_MAP_TAG:
                util.validate_node(tf, "custom rule", kind=yaml.MappingNode)
                tf_d = ctx.assemble_dict_keys(tf)
                fn = tf_d["fn"].value
                normalize = True
                if "normalize" in tf_d:
                    normalize = ctx.evaluate(tf_d["normalize"])
            else:
                raise util.Y4Error(f"Cannot define custom rule using {tf.tag}")
            yield (tag, CustomRule(fn, normalize=normalize))
